#!/usr/bin/env python

from stat import *
from copy import deepcopy, copy
from threading import Thread, Lock
from pysqlite2 import dbapi2 as db
from GuiUtils import _logerr, _logwarn, _loginfo, natsorted, thousands
from cherrypy import expose, HTTPError, tree, response, engine, thread_data, log
from cherrypy.lib.static import serve_file
from struct import pack, unpack
from cStringIO import StringIO
import libVisDQMServer as Accelerator
import os, re, time, socket, shutil, tempfile, md5, cgi
from PIL import Image, ImageEnhance

# --------------------------------------------------------------------
# Validate common DQM image request options.
RX_OPT_FLOAT = re.compile(r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)?$")
RX_SAFE_PATH = re.compile(r"^[A-Za-z0-9][-A-Za-z0-9_]*(?:\.(?:root|zip|ig))?$")

def _checkCommonImageOptions(options):
  return (options.get('w', '').isdigit()
          and options.get('h', '').isdigit()
          and options.get('xtype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('xmin', ''))
          and re.match(RX_OPT_FLOAT, options.get('xmax', ''))
          and options.get('ytype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('ymin', ''))
          and re.match(RX_OPT_FLOAT, options.get('ymax', ''))
          and options.get('ytype', '') in ('', 'lin', 'log')
          and re.match(RX_OPT_FLOAT, options.get('ymin', ''))
          and re.match(RX_OPT_FLOAT, options.get('ymax', ''))
          and re.match(r"^[A-Za-z ]*$", options.get('drawopts', ''))
          and options.get('ref', '') in ('', 'object',
                                         'reference',
                                         'overlay'))

# Utility to fetch an image from visDQMRender process.  Validates the
# image request options and builds a spec string to send to the
# backend render process.  Waits for the backend to reply with an
# image and returns either (MimeType, ImageData) tuple on success or
# (None, None) on failure.
#
# The remaining arguments should be the object name, the streamers
# and the streamed object data, each a string.  The streamers should
# be serialised ROOT TStreamerInfo required to deserialise the data
# (for ROOT objects) or empty (for scalar values).  The data should
# be either serialised ROOT objects or the scalar value as a string.
#
# The call will wait as long as the render process takes to reply.  A
# persistent per-thread socket is used to talk to the render process.
#
# PIL is used to scale down images that would be very small, as ROOT
# is not good at drawing small anti-aliased images.
#
# A cache of recent identical images is maintained, keyed by the spec
# and serialised object data, to avoid unnecessarily re-rendering the
# frequently accessed images.
#
# thread_data.reqsocket
#               Per server thread socket to the backend for image requests.
IMAGE_CACHE_LIST = []
IMAGE_CACHE_MAP = {}
IMAGE_CACHE_LOCK = Lock()
def ROOTImage(sockargs, options, path, streamers, data):
  if not _checkCommonImageOptions(options):
    return (None, None)

  spec = ""
  resize = None
  MIN_W, MIN_H = 532, 400
  for opt in ('drawopts', 'ref',
              'xtype', 'xmin', 'xmax',
              'ytype', 'ymin', 'ymax',
              'ztype', 'zmin', 'zmax'):
    if options.get(opt, '') != '':
      spec += "%s%s=%s" % ((spec and ";") or "", opt, options[opt])

  # If the image is small, generate a larger image in the same aspect
  # ratio and use PIL to produce a reduced size version.  ROOT itself
  # generates badly anti-aliased output.
  if 'w' in options and 'h' in options \
     and options['w'].isdigit() and options['h'].isdigit():
    w, h = int(options['w']), int(options['h'])
    if w > 0 and h > 0 and (w < MIN_W or h < MIN_H):
      resize = (w, h)
      aspect = float(w)/float(h)
      w = max(w, MIN_W)
      spec += "%sw=%d;h=%d" % ((spec and ";") or "", w, w/aspect)

  if not resize:
    for opt in ('w', 'h'):
      if options.get(opt, '') != '':
        spec += "%s%s=%s" % ((spec and ";") or "", opt, options[opt])

  # See if we have this image already cached.  If found, move the
  # image into the beginning of the cache (LRU ordering).
  imgtype = None
  imgdata = None
  imgkey = "%s:%s:%s" % (md5.new(spec).hexdigest(),
			 md5.new(data).hexdigest(),
			 (data == "" and path) or "")
  IMAGE_CACHE_LOCK.acquire()
  cached = IMAGE_CACHE_MAP.get(imgkey, None)
  if cached:
    imgtype = cached['type']
    imgdata = cached['image']
    IMAGE_CACHE_LIST.remove(cached)
    IMAGE_CACHE_LIST.insert(0, cached)
  IMAGE_CACHE_LOCK.release()

  # Now try to get the image from the backend.  Make a few tries
  # in case the socket has become invalid, or the backend dies
  # while we try.
  tryit = 3
  while imgtype == None and tryit > 0:
    tryit -= 1
    try:
      s = getattr(thread_data, 'reqsocket', None)
      if not s:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(sockargs)
        setattr(thread_data, 'reqsocket', s)

      payload = [path, spec, streamers, data]
      paylen = [len(i) for i in payload]
      nlen = 8 + 4 * len(paylen) + sum(paylen)
      payfmt = '%dI' % len(payload)
      s.send(pack('2I', nlen, 4))
      s.send(pack(payfmt, *paylen))
      s.send("".join(i for i in payload))

      reply = ""
      while s:
        part = s.recv(1048576)
	tryit = 0
        if len(part) == 0:
          s = None
          setattr(thread_data, 'reqsocket', s)

        reply += part
        if len(reply) < 8:
          continue

        (length, type) = unpack('2I', reply[0:8])
        if len(reply) < length:
          continue

        assert(len(reply) == length)
        assert(type == 105)

        imgdata = reply[8:]
        imgtype = "image/png"
        newitem = { 'key': imgkey, 'type': imgtype, 'image': imgdata }
	IMAGE_CACHE_LOCK.acquire()
	if imgkey not in IMAGE_CACHE_MAP:
          IMAGE_CACHE_MAP[imgkey] = newitem
          IMAGE_CACHE_LIST.insert(0, newitem)
	  while len(IMAGE_CACHE_LIST) > 5000:
            olditem = IMAGE_CACHE_LIST.pop()
	    del IMAGE_CACHE_MAP[olditem['key']]
	IMAGE_CACHE_LOCK.release()
	break

    except Exception, e:
      _logerr("failed to request image '%s' from %s as %s %d: %s"
	      % (path, sockargs, spec, len(streamers)+len(data), str(e)))
      setattr(thread_data, 'reqsocket', None)

  if imgtype and resize:
    try:
      out = StringIO()
      ImageEnhance.Sharpness(Image.open(StringIO(imgdata))
			     .resize(resize, Image.ANTIALIAS)) \
	.enhance(1.5).save(out, "png")
      imgdata = out.getvalue()
    except Exception, e:
      _logerr("failed to down scale image: %s" % e)

  return (imgtype, imgdata)

# --------------------------------------------------------------------
# Utility type for providing layouts from python configuration files.
class DQMItem:
  def __init__(self, layout=None):
    self.layout = layout

# --------------------------------------------------------------------
# DQM extension to manage DQM file uploads.
class DQMFileAccess:
  STATUS_OK               = 100 # Requested operation succeeded
  STATUS_BAD_REQUEST      = 200 # The request is malformed
  STATUS_ERROR_PARAMETER  = 300 # Request parameter value is unacceptable.
  STATUS_ERROR_EXISTS     = 301 # Cannot overwrite an existing object.
  STATUS_ERROR_NOT_EXISTS = 302 # Requested file does not exist.
  STATUS_FAIL_EXECUTE     = 400 # Failed to execute the request.

  def refresh(self, *args): pass
  def __init__(self, server, aclfile, datadir, roots):
    self.lock = Lock()
    self.server = server
    self.datadir = datadir
    self.roots = roots
    if datadir and not os.path.exists(datadir):
      os.makedirs(datadir)
    tree.mount(self, script_name=server.baseUrl + "/data",
               config={"/": {'request.show_tracebacks': False}})

  # Set response headers to indicate our status data.
  def _status(self, code, message, detail=None):
    response.headers['dqm-status-code'] = str(code)
    response.headers['dqm-status-message'] = message
    response.headers['dqm-status-detail'] = detail

  # Set response headers to indicate an error, then get out.
  def _error(self, code, message, detail=None):
    _logerr("code=%d, message=%s, detail=%s" % (code, message, detail))
    self._status(code, message, detail)
    raise HTTPError(500, message)

  # Check that a required parameter has been given just once,
  # and the value matches the given regular expression.
  def _check(self, name, arg, rx):
    if not arg or not isinstance(arg, str):
      self._error(self.STATUS_BAD_REQUEST,
                  "Incorrect or missing %s parameter" % name,
                  "Must provide single argument")
    if not re.match(rx, arg):
      self._error(self.STATUS_BAD_REQUEST,
                  "Malformed %s argument" % name,
                  "Argument must match regular expression '%s'" % rx)

  # ------------------------------------------------------------------
  # Store a file to the server.  Validates all the parameters, then
  # attempts to save the file safely.  Sets headers in the response
  # to indicate what happened, either an error or success.
  # FIXME: Verify the request was submitted securely and is permitted.
  # FIXME: Determine producer from certificate information.
  @expose
  def put(self,
          producer = None,
          step = None,
          workflow = None,
          filename = None,
          mssname = None,
          size = None,
          checksum = None,
          file = None,
          *args,
          **kwargs):
    # Bail out if upload is not supported.
    if not self.datadir:
      raise HTTPError(404, "Not found")

    # Basic argument validation.
    if file == None \
       or not getattr(file, "file", None) \
       or not getattr(file, "filename", None):
      self._error(self.STATUS_BAD_REQUEST,
                  "Incorrect or missing file argument",
                  "Must provide a single file-type argument")

    self._check("producer", producer,      r"^[A-Za-z0-9]+$")
    self._check("step",     step,          r"^[-A-Za-z0-9]+$")
    self._check("workflow", workflow,      r"^(/[-A-Za-z0-9_]+){3}$")
    self._check("filename", file.filename, r"^[-A-Za-z0-9_]+\.root$")
    self._check("mssname",  mssname,       r"^[-A-Za-z0-9_./:=?]+$")
    self._check("size",     size,          r"^\d+$")
    self._check("checksum", checksum,      r"^(md5:[A-Za-z0-9]+|crc32:\d+)$")

    # Argument value validation.
    if producer not in ("ProdSys",):
      self._error(self.STATUS_ERROR_PARAMETER,
                  "This producer is not authorised")
    if step not in ("Pass-1", "PromptReco"):
      self._error(self.STATUS_ERROR_PARAMETER,
                  "This step is not authorised")

    m = re.match(r"^(DQM)_V\d+_R(\d+)(__.*)\.root", file.filename)
    if not m:
      self._error(self.STATUS_ERROR_PARAMETER,
                  "File name does not match the expected convention")

    size = int(size)
    runnr = int(m.groups()[1])
    wflow = m.groups()[2]
    if workflow.replace("/", "__") != wflow:
      self._error(self.STATUS_ERROR_PARAMETER,
                  "File name does not match the workflow name")
    stream = workflow.split("/")[1]

    # Determine where we would save this file.
    dir = "%s/%s/%s/%s/%03d/%03d/%03d" % \
          (self.datadir, producer, step, stream,
           (runnr / 1000000) % 1000, (runnr / 1000) % 1000, runnr % 1000)
    fname = "%s/%s" % (dir, file.filename)
    tmp = None

    if fname.find("/../") >= 0 or fname.find("/./") >= 0:
      self._error(self.STATUS_ERROR_PARAMETER, "Invalid file path name",
                  "Path name must not refer to '.' or '..'")

    if mssname.find("/../") >= 0 or mssname.find("/./") >= 0:
      self._error(self.STATUS_ERROR_PARAMETER, "Invalid file path name",
                  "Path name must not refer to '.' or '..'")

    if os.path.exists(fname) or os.path.exists(fname + ".origin"):
      self._error(self.STATUS_ERROR_EXISTS,
                  "File exists, cannot overwrite", fname)

    # Try saving the file safely.  First we write the file to a
    # temporary directory, with a fake name, to the same directory as
    # the final file would be.  We then move it in place atomically
    # and leave a marker to indicate where the MSS replica of the file
    # can be found. If anything goes wrong, clean up so the upload can
    # be re-attempted later without tripping over the checks above.
    try:
      self.lock.acquire()
      if not os.path.exists(dir):
        os.makedirs(dir)
      (fd, tmp) = tempfile.mkstemp(".upload", "", dir)
      nsaved = 0
      while True:
        data = file.file.read(32*1024)
        if not data:
          break
        os.write(fd, data)
        nsaved += len(data)
      os.close(fd)
      os.chmod(tmp, 0644)
      if nsaved != size:
        self._error(self.STATUS_FAIL_EXECUTE,
                    "Failed to save file data",
                    "Wrote %d bytes, expected to write %d" % (nsaved, size))
      shutil.move(tmp, fname)
      open(fname + ".origin", "w").write("%s %d %s\n" % (checksum, size, mssname))
      self.lock.release()

    except Exception, e:
      self.lock.release()
      if os.path.exists(fname):
        os.remove(fname)
      if os.path.exists(fname + ".origin"):
        os.remove(fname + ".origin")
      if tmp and os.path.exists(tmp):
        os.remove(tmp)
      if isinstance(e, HTTPError):
        raise e
      self._error(self.STATUS_FAIL_EXECUTE,
                  "Failed to save file data",
                  str(e).replace("\n", "; "))

    # Indicate success.
    self._status(self.STATUS_OK, "File saved",
                 "Wrote %d bytes, reference in %s" % (nsaved, mssname))
    _loginfo("saved file path=%s msspath=%s size=%d mssname=%s"
             % (fname, tmp, size, mssname))
    return "Thanks.\n"

  # ------------------------------------------------------------------
  # Retrieve files from the server.  Pretends to be somewhat like the
  # apache mod_dir file browsing scheme.  Serves only files with valid
  # and safe path references.  FIXME: Bandwidth limiting and ACLs?
  @expose
  def browse(self, *path, **kwargs):
    rooturl = self.server.baseUrl + "/data/browse"
    if not self.roots or len(self.roots) == 0:
      raise HTTPError(404, "Not found")

    for x in path:
      if not re.match(RX_SAFE_PATH, x):
	raise HTTPError(404, "Not found")

    # If no path was given, provide list of top level roots.
    if len(path) == 0:
      return "<html><head><title>Index of %(index)s</title></head><body>" \
             "<h1>%(index)s</h1><table border='0' valign='top' cellpadding='5'>" \
	     "%(files)s</table></body></html>" % \
             { 'index': cgi.escape("DQM files"),
	       'files': "\n".join(["<tr><td><a href='%(root)s/%(name)s'>%(name)s"
				   "</a></td><td>&nbsp;</td><td>&nbsp;</td>"
				   % { 'root': rooturl, 'name': cgi.escape(x) }
				   for x in sorted(self.roots.keys())]) }

    # We have a path, check under the given root
    root = path[0]
    if root not in self.roots:
      raise HTTPError(404, "Not found")

    pathname = "/".join([self.roots[root]] + list(path[1:]))
    if not os.path.exists(pathname):
      raise HTTPError(404, "Not found")

    if os.path.isdir(pathname):
      files = [(x, pathname + "/" + x)
	       for x in os.listdir(pathname)
	       if re.match(RX_SAFE_PATH, x)]
      files = [(x[0], x[1], os.path.isdir(x[1]), os.stat(x[1]))
	       for x in files if not os.path.islink(x[1])]
      return "<html><head><title>Index of %(index)s</title></head><body>" \
             "<h1>%(index)s</h1><p style='padding-left:5px'>" \
	     "<a href='%(root)s/%(parent)s'>Up</a></p>" \
	     "<table border='0' valign='top' cellpadding='5'>" \
	     "%(files)s</table></body></html>" % \
             { 'index': cgi.escape("/".join(path)),
	       'parent': cgi.escape("/".join(path[:-1])), 'root': rooturl,
	       'files': "\n".join(["<tr><td><a href='%(root)s/%(path)s/%(name)s'>"
				   "%(name)s%(slash)s</a></td><td>%(size)s</td>"
				   "<td>%(mtime)s UTC</td></tr>"
				   % { 'root': rooturl,
				       'path': cgi.escape("/".join(path)),
				       'name': cgi.escape(x[0]),
				       'slash': (x[2] and '/') or "",
				       'size': (x[2] and '-') or x[3].st_size,
				       'mtime': time.strftime("%Y-%m-%d %H:%M:%S",
						  time.gmtime(x[3].st_mtime)) }
				   for x in sorted(files)[::-1]]) }
    else:
      return serve_file(pathname, content_type="application/octet-stream")

# --------------------------------------------------------------------
# Source for plotting non-existent DQM objects.  Feeds the names of
# missing objects the visDQMRender for "missing in action" image.
#
# .port         The port at which visDQMRender is listening.
class DQMUnknownSource(Accelerator.DQMUnknownSource):
  # Notify plot fairy the subdirectory where we serve images.
  port = 9091

  def __init__(self, server, statedir, port = None):
    Accelerator.DQMUnknownSource.__init__(self)
    if port: self.port = port

  # Generate a "missing in action" image for an unknown object.
  def plot(self, *path, **options):
    return ROOTImage(("localhost", self.port), options, path[-1], "", "")

# --------------------------------------------------------------------
# DQM data source which provides layout content from python
# configuration files.  There is no backend process attached to this
# data source.
#
# Takes a list of python configuration filenames as the configuration
# parameter.  The files contain code to initialise a static structure
# of layout definitions.
#
# Whenever the source is prompted to update its data, it first reviews
# the list of layout configuration files.  Changed files are reloaded;
# if the file fails to load it is skipped, although layout
# registrations up to the point of error are included.
#
# Once the layout definitions are up-to-date, simply returns the list
# of known layouts.  No checks are made to ensure the listed monitor
# elements actually exist; non-existent ones will simply generate a
# warning when rendered.
#
# This source does not produce plots.  Plots are provided by the
# sources which provide the objects the layouts refer to.  All the
# real functionality is implemented in the C++ layer.
class DQMLayoutSource(Accelerator.DQMLayoutSource):
  def __init__(self, server, statedir, *filenames):
    # Filter only existing files.
    files = [f for f in filenames if os.path.exists(f)]

    # Pass the filtered list to the implementation.
    Accelerator.DQMLayoutSource.__init__(self, files)

    # Add file checksums to server configuration.
    for f in files:
      server._addChecksum(None, f, open(f).read())

# --------------------------------------------------------------------
# DQM data source providing content from DQM data delivered over
# network from upstream "DQMCollector".
#
# Takes as parameter the options for the "visDQMRender" process.  This
# GUI server side object decodes those arguments and determines the
# host and port to connect to reach the backend processes.
#
# All the monitoring and processing is done in the C++ accelerator
# class this python class inherits from.  The python layer is just
# a glue layer to pass data between the two.
#
# This source works with visDQMRender to make object images: we get
# the object data via DQMCollector and pass it to visDQMRender which
# returns back the PNG image.  The image requests block until the
# render process returns the image data.  Each HTTP server thread
# has a semi-persistent socket to the render process; cf. ROOTImage.
# Note that there is no time out on retrieving image data from the
# upstream source, the DQM network layer will automatically time out
# after a while.  If the image rendering gets stuck, the process will
# automatically commit a suicide and restart.
#
# .opts         Configuration parameters
#   dataset     - Dataset name for online datasets (fixed)
#   verbose     - Flag to cause DQM net to be chatty or quiet (fixed)
#   dqmhost     - DQMCollector host name, taken from --collector option.
#   dqmport     - DQMCollector port, taken from --collector option.
#   plotport    - Port at which visDQMRender is listening, taken from --listen option.
class DQMLiveSource(Accelerator.DQMLiveSource):
  # Start off the live backend and hook up to server transitions.
  def __init__(self, server, statedir, collector, *params):
    self.opts = { "dataset": "/Global/Online/ALL",
                  "verbose": False,
                  "dqmhost": "localhost",
                  "dqmport": 9090,
                  "plotport": 9091 }

    m = re.match(r"((\S+):)?(\d+)", collector)
    if m:
      if m.group(2) != "":
        self.opts['dqmhost'] = m.group(2)
      self.opts['dqmport'] = int(m.group(3))

    for arg in params:
      m = re.match(r"--listen (\d+)", arg)
      if m:
        self.opts['plotport'] = int(m.group(1))
        continue

    Accelerator.DQMLiveSource.__init__(self, server, self.opts)
    engine.on_stop_engine_list.append(lambda *args: self._stop())
    engine.on_stop_thread_list.append(lambda *args: setattr(thread_data, 'reqsocket', None))

  # Generate an object image given an object name and render
  # options.  See ROOTImage for details about image generation.
  def plot(self, *path, **options):
    objname = "/".join(path)
    (streamers, data) = self._data(objname)
    return ROOTImage(("localhost", self.opts["plotport"]),
		     options, objname, streamers, data)

# --------------------------------------------------------------------
# DQM data source providing content from archived DQM data files.
# All the real functionality is implemented in the C++ layer.  This
# source works with visDQMRender to make object images: we ship the
# streamer info and streamed objects to the backend and it returns
# back the PNG image.
#
# .opts         Configuration parameters
#   plotport    - Port at which visDQMRender is listening.
#   rxonline    - Regexp to recognise online dataset name.
#   index       - Path to the index directory.
class DQMArchiveSource(Accelerator.DQMArchiveSource):
  def __init__(self, server, statedir, indexdir, rxonline, *params):
    self.opts = { "index": indexdir, "rxonline": rxonline, "plotport": 9092 }
    for arg in params:
      m = re.match(r"--listen (\d+)", arg)
      if m:
        self.opts['plotport'] = int(m.group(1))
        continue

    Accelerator.DQMArchiveSource.__init__(self, server, self.opts)
    engine.on_stop_thread_list.append(lambda *args: self._stop())
    engine.on_stop_thread_list.append(lambda *args: setattr(thread_data, 'reqsocket', None))

  # Generate an object image given an object type ('scalar' or
  # 'rootobj'), the run number, dataset path, object name and render
  # options.  See ROOTImage for details about image generation.
  def plot(self, objtype, sampletype, runnr, dsPrim, dsWflow, dsTier, *path, **options):
    if not hasattr(Accelerator.objtype, objtype):
      raise HTTPError(500, "Incorrect run parameter")
    if not hasattr(Accelerator.sampletype, sampletype):
      raise HTTPError(500, "Incorrect run parameter")
    objname = "/".join(path)
    dataset = "/".join(('', dsPrim, dsWflow, dsTier))
    (streamers, data) = self._data(getattr(Accelerator.objtype, objtype),
    				   getattr(Accelerator.sampletype, sampletype),
				   int(runnr), dataset, objname)
    return ROOTImage(("localhost", self.opts["plotport"]), options, objname, streamers, data)

# --------------------------------------------------------------------
# .sessiondef
# .gui
# .rank
# .category
# .name
# .match
# .warnings         Recent warnings printed to the logs.

class DQMWorkspace:
  sessiondef = { 'dqm.sample.vary':    "any",
                 'dqm.sample.order':   "dataset",
                 'dqm.sample.prevws':  "",
                 'dqm.sample.pattern': "",
                 #'dqm.sample.type':    0,
                 #'dqm.sample.runnr':   0,
                 #'dqm.sample.dataset': "",
                 'dqm.play.prevws':    "",
                 'dqm.play.interval':  5,
                 'dqm.play.pos':       0,
                 #'dqm.panel.help.show': 1,
                 #'dqm.panel.help.x':  -1,
                 #'dqm.panel.help.y':  -1,
                 #'dqm.panel.custom.show': 1,
                 #'dqm.panel.custom.x':  -1,
                 #'dqm.panel.custom.y':  -1,
		 'dqm.qplot': 	       "",
                 'dqm.search':         "",
                 'dqm.filter':         "all",
                 'dqm.reference':      "customise",
                 'dqm.submenu':	       "data",
                 'dqm.size':           "M",
                 'dqm.root':           {},
                 'dqm.focus':          {},
                 'dqm.drawopts':       {},
                 'dqm.myobjs':         {} }

  def __init__(self, gui, rank, category, name):
    #gui._addCSSFragment("%s/fonts/fonts.css" % gui._yui)
    gui._addCSSFragment("%s/container/assets/skins/sam/container.css" % gui._yui)
    gui._addCSSFragment("%s/resize/assets/skins/sam/resize.css" % gui._yui)
    gui._addJSFragment("%s/utilities/utilities.js" % gui._yui)
    gui._addJSFragment("%s/container/container.js" % gui._yui)
    gui._addJSFragment("%s/resize/resize.js" % gui._yui)

    gui.css = [x for x in gui.css if x[0] != '%s/style-core.css' % gui.contentpath]
    gui._addCSSFragment("%s/style-dqm.css" % gui.contentpath)
    for p in ("canvas", "header", "quality", "sample", "summary", "play"):
      gui._addJSFragment("%s/gui-dqm-%s.js" % (gui.contentpath, p))

    self.gui       = gui
    self.rank      = rank
    self.category  = category
    self.name      = name
    self.warnings  = {}

  # Customise server on start-up. Force hidden internal workspaces.
  def customise(self):
    if "Sample" not in (w.name for w in self.gui.workspaces):
      self.gui.workspaces.append(DQMSampleWorkspace(self.gui))

    if "Play" not in (w.name for w in self.gui.workspaces):
      self.gui.workspaces.append(DQMPlayWorkspace(self.gui))

  # Initialise a new session.
  def initialiseSession(self, session):
    for (var, value) in self.sessiondef.iteritems():
      if var not in session:
        session[var] = deepcopy(value)

  # Make sure the accessed object name is valid.
  def _checkObjectName(self, o):
    if not re.match(r"^[-+=_()/# A-Za-z0-9]*$", o):
      if o not in self.warnings:
        _logwarn("attempt to access unsafe object or version '%s'" % o)
	self.warnings[o] = 1
      raise HTTPError(500, "Attempt to access unsafe object")
    return o

  # Set session parameters, checking them for validity.
  def _set(self, session,
           samplevary = None,
           sampleorder = None,
           sampleprevws = None,
           samplepat = None,
           sampletype = None,
           dataset = None,
           runnr = None,
	   qplot = None,
           filter = None,
           reference = None,
           search = None,
           submenu = None,
           size = None,
           playprevws = None,
           playinterval = None,
           playpos = None,
           root = None,
           focus = None,
           reset = None,
           add = None,
           **kwargs):

    if samplevary != None:
      if not isinstance(samplevary, str) \
         or samplevary not in ("any", "run", "dataset"):
        raise HTTPError(500, "Incorrect sample vary parameter")
      session['dqm.sample.vary'] = samplevary

    if sampleorder != None:
      if not isinstance(sampleorder, str) \
         or sampleorder not in ("run", "dataset"):
        raise HTTPError(500, "Incorrect sample order parameter")
      session['dqm.sample.order'] = sampleorder

    if sampleprevws != None:
      if not isinstance(sampleprevws, str) \
         or sampleprevws not in (x.name for x in self.gui.workspaces):
        raise HTTPError(500, "Incorrect sample previous workspace parameter")
      session['dqm.sample.prevws'] = sampleprevws

    if samplepat != None:
      if not isinstance(samplepat, str):
        raise HTTPError(500, "Incorrect sample pattern parameter")
      session['dqm.sample.pattern'] = samplepat

    if sampletype != None:
      if not isinstance(sampletype, str):
        raise HTTPError(500, "Incorrect sample type parameter")
      elif hasattr(Accelerator.sampletype, sampletype):
        session['dqm.sample.type'] = int(getattr(Accelerator.sampletype, sampletype))
      else:
        raise HTTPError(500, "Incorrect sample type parameter")

    if dataset != None:
      if not isinstance(dataset, str):
        raise HTTPError(500, "Incorrect dataset parameter")
      session['dqm.sample.dataset'] = dataset

    if runnr != None:
      if not isinstance(runnr, str):
        raise HTTPError(500, "Incorrect run parameter")
      elif not re.match(r"^\d+$", runnr):
        raise HTTPError(500, "Incorrect run parameter")
      else:
        session['dqm.sample.runnr'] = int(runnr)

    if qplot != None:
      if not isinstance(qplot, str):
        raise HTTPError(500, "Incorrect quality plot parameter")
      if not re.match(r"^[A-Za-z.\d]*$", qplot):
        raise HTTPError(500, "Invalid quality plot parameter")
      session['dqm.qplot'] = qplot

    if filter != None:
      if not isinstance(filter, str) \
         or filter not in ("all", "alarms", "nonalarms"):
        raise HTTPError(500, "Incorrect filter parameter")
      session['dqm.filter'] = filter

    if reference != None:
      if not isinstance(reference, str) \
         or reference not in ("overlay", "on-side", "customise"):
        raise HTTPError(500, "Incorrect reference parameter")
      session['dqm.reference'] = reference

    if search != None:
      if not isinstance(search, str):
        raise HTTPError(500, "Incorrect search parameter")
      session['dqm.search'] = search

    if submenu != None:
      if not isinstance(submenu, str) \
         or submenu not in ("data", "service", "workspace"):
        raise HTTPError(500, "Incorrect submenu parameter")
      session['dqm.submenu'] = submenu

    if size != None:
      if not isinstance(size, str) \
         or size not in ("XS", "S", "M", "L", "XL"):
        raise HTTPError(500, "Incorrect size parameter")
      session['dqm.size'] = size

    if playprevws != None:
      if not isinstance(playprevws, str) \
         or playprevws not in (x.name for x in self.gui.workspaces):
        raise HTTPError(500, "Incorrect play previous workspace parameter")
      session['dqm.play.prevws'] = playprevws

    if playinterval != None:
      if not isinstance(playinterval, str) \
         or not re.match(r"^\d+$", playinterval):
        raise HTTPError(500, "Incorrect play interval parameter")
      playinterval = int(playinterval)
      if playinterval == 0:
        playinterval = self.sessiondef['dqm.play.interval']
      elif playinterval < 2:
        playinterval = 2
      elif playinterval > 99:
        playinterval = 99
      session['dqm.play.interval'] = playinterval

    if playpos != None:
      if not isinstance(playpos, str) \
         or not re.match(r"^\d+$", playpos):
        raise HTTPError(500, "Incorrect play position parameter")
      session['dqm.play.pos'] = int(playpos)

    if root != None:
      if not isinstance(root, str):
        raise HTTPError(500, "Incorrect root parameter")
      session['dqm.root'][self.name] = self._checkObjectName(root)

    if focus != None:
      if not isinstance(focus, str):
        raise HTTPError(500, "Incorrect focus parameter")
      elif focus == "":
        if self.name in session['dqm.focus']:
          del session['dqm.focus'][self.name]
      else:
        session['dqm.focus'][self.name] = self._checkObjectName(focus)

    if reset == "yes":
      session['dqm.myobjs'][self.name] = []

    if add != None:
      session['dqm.myobjs'][self.name].extend \
        ([self._checkObjectName(x)
          for x in ((type(add) == str and [ add ]) or add)])

  # -----------------------------------------------------------------
  # Initialise a new session with the "start" URL.
  def start(self, session, **kwargs):
    self._set(session, **kwargs)

  # -----------------------------------------------------------------
  # Return DQM-specific index page.
  def sessionIndex(self, session, *args, **kwargs):
    return self.gui._templatePage("index-dqm", {
        'TITLE'      : self.gui.title,
        'SESSION_ID' : session['core.name'],
        'HOSTNAME'   : socket.gethostname(),
        'ROOTPATH'   : self.gui.baseUrl
      })

  # -----------------------------------------------------------------
  # Switch session to another workspace.  Resets to data sub-menu
  # then re-uses GUI core workspace method.
  def sessionWorkspace(self, session, *args, **kwargs):
    self._set(session, submenu = "data")
    return self.gui.sessionWorkspace(session, *args, **kwargs)

  # Set quality plot selection.  For notification only.
  def sessionSetQualityPlot(self, session, *args, **kwargs):
    self._set(session, qplot = kwargs.get("n", ""))
    self.gui._saveSession(session)
    return "Ta."

  # Start choosing another sample.  This switches to a hidden internal
  # auto-generated workspace, remembering where the user was previously.
  def sessionChooseSample(self, session, *args, **kwargs):
    self._set(session,
              samplevary = kwargs.get("vary", "any"),
              sampleorder = kwargs.get("order", "dataset"),
              sampleprevws = session['core.workspace'])
    return self.gui.sessionWorkspace(session, name = "sample")

  # Set all/alarm/non-alarm filter.
  def sessionSetFilter(self, session, *args, **kwargs):
    self._set(session, filter = kwargs.get("n", "all"))
    self.gui._saveSession(session)
    return self._state(session)

  # Set global reference display override.
  def sessionSetReference(self, session, *args, **kwargs):
    self._set(session, reference = kwargs.get("n", "customise"))
    self.gui._saveSession(session)
    return self._state(session)

  # Set sub-menu.  For notification only.
  def sessionSetSubMenu(self, session, *args, **kwargs):
    self._set(session, submenu = kwargs.get("n", "data"))
    self.gui._saveSession(session)
    return "Ta."

  # Set content search pattern.
  def sessionSetSearch(self, session, *args, **kwargs):
    self._set(session, search = kwargs.get("rx", ""))
    self.gui._saveSession(session)
    return self._state(session)

  # Set the size of canvas objects.
  def sessionSetSize(self, session, *args, **kwargs):
    self._set(session, size = kwargs.get("sz", None))
    self.gui._saveSession(session)
    return self._state(session)

  # Turn play mode on or off.
  def sessionPlay(self, session, *args, **kwargs):
    self._set(session, playprevws = session['core.workspace'])
    return self.gui.sessionWorkspace(session, name = "play")

  # Change the root shown in the selector menu.
  def sessionSetRoot(self, session, *args, **kwargs):
    self._set(session, root = kwargs.get("n", ""))
    if self.name in session['dqm.focus']:
      del session['dqm.focus'][self.name]
    self.gui._saveSession(session)
    return self._state(session)

  # Change the selected object.
  def sessionSetFocus(self, session, *args, **kwargs):
    name = kwargs.get("n", None)
    if name == None:
      if self.name in session['dqm.focus']:
        del session['dqm.focus'][self.name]
    else:
      self._set(session, focus = name)
    self.gui._saveSession(session)
    return "Ta."

# --------------------------------------------------------------------
class DQMSummaryWorkspace(Accelerator.DQMSummaryWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name):
    Accelerator.DQMSummaryWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

# --------------------------------------------------------------------
class DQMQualityWorkspace(Accelerator.DQMQualityWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name):
    Accelerator.DQMQualityWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

# --------------------------------------------------------------------
class DQMContentWorkspace(Accelerator.DQMContentWorkspace, DQMWorkspace):
  def __init__(self, gui, rank, category, name, match, *content):
    Accelerator.DQMContentWorkspace.__init__(self, gui, name, match)
    DQMWorkspace.__init__(self, gui, rank, category, name)
    self.content = content

  # Initialise a new session.
  def initialiseSession(self, session):
    DQMWorkspace.initialiseSession(self, session)
    if self.name not in session['dqm.myobjs'] and len(self.content):
      session['dqm.myobjs'][self.name] = deepcopy(self.content)
      session['dqm.root'][self.name] = "Quick collection"

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

  # -----------------------------------------------------------------
  # Reset the canvas back to the workspace default.
  def sessionResetWorkspace(self, session, *args, **kwargs):
    session['dqm.play.prevws'] = ""
    session['dqm.play.pos'] = 0
    session['dqm.size'] = self.sessiondef['dqm.size']
    for key in ('dqm.focus', 'dqm.drawopts', 'dqm.myobjs'):
      if self.name in session[key]:
        del session[key][self.name]
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session)

  # -----------------------------------------------------------------
  # Change image parameters.
  def sessionAlterImage(self, session, *args, **kwargs):
    obj = kwargs['o']
    opt = kwargs['opt']
    val = kwargs['value']
    if not isinstance(obj, str):
      raise HTTPError(500, "Image operation without an image")
    if opt == "withref":
      if val == "":
        val == "no"
      if val not in ("yes", "no"):
        raise HTTPError(500, "Unrecognised reference setting")
    elif opt == "drawopts":
      if not re.match(r"^[A-Za-z,]*$", val):
        raise HTTPError(500, "Unknown draw options")
    elif opt in ("xtype", "ytype", "ztype"):
      if val == "":
	val = "lin"
      if val not in ("lin", "log"):
        raise HTTPError(500, "Unrecognised axis type")
    elif opt in ("xmin", "xmax", "ymin", "ymax", "zmin", "zmax"):
      if not re.match(r"^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)?$", val):
        raise HTTPError(500, "Unrecognised axis limit")
    else:
      raise HTTPError(500, "Attempt to set unknown image option")

    opts = session['dqm.drawopts'].setdefault(self.name, {}).setdefault(obj, {})
    if val == '':
      if opt in opts:
        del opts[opt]
      if len(opts) == 0:
        del session['dqm.drawopts'][self.name][obj]
      if len(session['dqm.drawopts'][self.name]) == 0:
        del session['dqm.drawopts'][self.name]
    else:
      opts[opt] = val

    self.gui._saveSession(session)
    return self._state(session)

  # -----------------------------------------------------------------
  # Change image parameters.
  def sessionPanel(self, session, *args, **kwargs):
    name = kwargs.get('n')
    show = kwargs.get('show', None)
    x = kwargs.get('x', None)
    y = kwargs.get('y', None)

    if not isinstance(name, str) or name not in ("help", "custom"):
      raise HTTPError(500, "Panel operation without a valid panel name")

    if show != None:
      if not isinstance(show, str) or show not in ("yes", "no"):
        raise HTTPError(500, "Invalid panel show parameter")
      session['dqm.panel.%s.show' % name] = show == "yes"

    if x != None:
      if not isinstance(x, str) or not re.match(r"^\d+$", x):
        raise HTTPError(500, "Invalid panel position parameter")
      session['dqm.panel.%s.x' % name] = int(x)

    if y != None:
      if not isinstance(y, str) or not re.match(r"^\d+$", y):
        raise HTTPError(500, "Invalid panel position parameter")
      session['dqm.panel.%s.y' % name] = int(y)

    self.gui._saveSession(session)
    return self._state(session)

# --------------------------------------------------------------------
class DQMShiftWorkspace(DQMContentWorkspace):
  SHIFT_ROOT = "00 Shift"

  def __init__(self, gui, rank, category, name):
    DQMContentWorkspace.__init__(self, gui, rank, category, name, "^")

  # Initialise a new session.
  def initialiseSession(self, session):
    DQMContentWorkspace.initialiseSession(self, session)
    session['dqm.root'][self.name] = self.SHIFT_ROOT

  # Override session state to fix root.
  def sessionState(self, session, *args, **kwargs):
    if not session['dqm.root'].get(self.name, "").startswith(self.SHIFT_ROOT):
      session['dqm.root'][self.name] = self.SHIFT_ROOT
      self.gui._saveSession(session)
    return DQMContentWorkspace.sessionState(self, session)

  # Change the root shown in the selector menu.
  def sessionSetRoot(self, session, *args, **kwargs):
    if kwargs.get("n", "").startswith(self.SHIFT_ROOT):
      return DQMContentWorkspace.sessionSetRoot(self, session, *args, **kwargs)
    else:
      return self._state(session)

# --------------------------------------------------------------------
class DQMSampleWorkspace(Accelerator.DQMSampleWorkspace, DQMWorkspace):
  def __init__(self, gui, rank = -1, category = "Internal", name = "Sample"):
    Accelerator.DQMSampleWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

  # Return back from sample selection without making changes.
  def sessionReturn(self, session, *args, **kwargs):
    prevws = session['dqm.sample.prevws']
    session['dqm.sample.prevws'] = ""
    return self.gui.sessionWorkspace(session, name = prevws)

  # Set the search pattern, order or parameter to vary.
  def sessionModify(self, session, *args, **kwargs):
    self._set(session,
	      samplevary = kwargs.get("vary", None),
	      sampleorder = kwargs.get("order", None),
	      samplepat = kwargs.get("pat", None));
    self.gui._saveSession(session)
    return self._state(session)

  # Select another sample.
  def sessionSelect(self, session, *args, **kwargs):
    self._set(session,
              sampletype = kwargs.get("type"),
              dataset = kwargs.get("dataset"),
              runnr = kwargs.get("runnr"));
    return self.sessionReturn(session)

# --------------------------------------------------------------------
class DQMPlayWorkspace(Accelerator.DQMPlayWorkspace, DQMWorkspace):
  def __init__(self, gui, rank = -1, category = "Internal", name = "Play"):
    Accelerator.DQMPlayWorkspace.__init__(self, gui, name)
    DQMWorkspace.__init__(self, gui, rank, category, name)

  # Return JSON object for the current session state.
  def sessionState(self, session, *args, **kwargs):
    return self._state(session)

  # Return back from sample selection without making changes.
  def sessionStop(self, session, *args, **kwargs):
    prevws = session['dqm.play.prevws']
    session['dqm.play.prevws'] = ""
    return self.gui.sessionWorkspace(session, name = prevws)

  # Set the search pattern.
  def sessionInterval(self, session, *args, **kwargs):
    self._set(session, playinterval = kwargs.get("v", self.sessiondef['dqm.play.interval']))
    self.gui._saveSession(session)
    return self._state(session)

  # Select another sample.
  def sessionStep(self, session, *args, **kwargs):
    self._set(session, playpos = kwargs.get("pos", 0))
    if kwargs.get("move", "no") == "yes":
      return self._state(session)
    else:
      return "Ta."
